import os
import time
import sys
import ipaddress
import socket

cache_dir = "/tmp/"
cache_life = 3600 * 24
dirty = False

valid_orgs = ['CS']

def fetch_subnets():
    import urllib.request
    import re
    import time
    import ipaddress

    pass #sys.stderr.write("DEBUG: Performing subnet fetch on subnet\n")

    url = "https://cs.uwaterloo.ca/cscf/internal/infrastructure/auto_generated/IST_data/networks.txt"
    resp = urllib.request.urlopen(url)
    line = ""
    subnets = {}
    subnets['timestamp'] = time.time()
    subnets['subnets'] = {}
    while True:
        try:
            try:
                line = resp.readline().decode('utf-8')
            except:
                pass #sys.stderr.write("WARNING: Failed to parse line as valid utf-8\n")
                continue

            if line == "":
                break

            try:
                address = re.split(r'\s+', line)[1]
                mask = re.split(r'\s+', line)[4]
                name = re.split(r'\s+', line)[0]
                organization = re.split(r'\s+', line)[3]
            except:
                pass #sys.stderr.write("WARNING: Unable to obtain address, mask, name or organization from:\n " + line + "\n")
                continue

            subnet = ipaddress.ip_network(address + mask)

            if( subnet.version == 4 ):
                if ( organization in valid_orgs ):
                    subnets['subnets'][address + mask] = {}
                    subnets['subnets'][address + mask]['name'] = name
                    subnets['subnets'][address + mask]['organization'] = organization
                else:
                    pass #sys.stderr.write("WARNING: " + address + mask + " Invalid organization " + organization + "\n")
                    continue
            else:
                pass #sys.stderr.write("WARNING: Not IPV4 subnet\n")
                continue

        except:
            raise

    global dirty
    dirty = True
    return subnets

def fetch_vlan(subnet):
    import urllib.request
    import re
    import time

    #sys.stderr.write("DEBUG: Performing vlan fetch on subnet " + subnet + "\n")

    #sys.stderr.write("DEBUG: URL: " + url + "\n")
    #url = "http://istns.uwaterloo.ca/onaquery.php?table=vlannames&ipsubnet=" + subnet.split('/',1)[0]
    url = "http://hp-ona.private.uwaterloo.ca/onaquery.php?table=vlannames&ipsubnet="  + subnet.split('/',1)[0]
    try:
        resp = urllib.request.urlopen(url)
        lines = resp.readlines()
    except:
        raise
    if ( lines[0].decode('utf-8').split('=')[0] == 'error' ):
        #sys.stderr.write("WARNING: Got error fetching vlan for subnet: " + subnet + "\n")
        return None

    vlan = {}
    try:
        vlan['device'] = lines[0].decode('utf-8').split('=')[1].rstrip()
    except:
        #sys.stderr.write("WARNING: Could not get device\n")
        return None
    try:
        vlan['name'] = lines[1].decode('utf-8').split('=')[1].rstrip()
    except:
        #sys.stderr.write("WARNING: Could not get name\n")
        return None
    try:
        vlan['id'] = lines[2].decode('utf-8').split('=')[1].rstrip()
    except:
        #sys.stderr.write("WARNING: Could not get id\n")
        return None
    try:
        vlan['ifndex'] = lines[3].decode('utf-8').split('=')[1].rstrip()
    except:
        #sys.stderr.write("WARNING: Could not get ifindex\n")
        return None
    try:
        vlan['taggedifndex'] = lines[4].decode('utf-8').split('=')[1].rstrip()
    except:
        #sys.stderr.write("WARNING: Could not get taggedifindex\n")
        return None
    try:
        vlan['ipaddress'] = lines[5].decode('utf-8').split('=')[1].rstrip()
    except:
        #sys.stderr.write("WARNING: Could not get taggedifindex\n")
        return None
    try:
        vlan['ipmask'] = lines[6].decode('utf-8').split('=')[1].rstrip()
    except:
        #sys.stderr.write("WARNING: Could not get ipmask\n")
        return None
    try:
        vlan['ipsubnet'] = lines[7].decode('utf-8').split('=')[1].rstrip()
    except:
        #sys.stderr.write("WARNING: Could not get ipsubnet\n")
        return None
    try:
        vlan['netlength'] = lines[8].decode('utf-8').split('=')[1].rstrip()
    except:
        #sys.stderr.write("WARNING: Could not get netlength\n")
        return None
    try:
        vlan['router'] = lines[9].decode('utf-8').split('=')[1].rstrip()
    except:
        #sys.stderr.write("WARNING: Could not get router\n")
        return None
    try:
        vlan['timestamp'] = time.time()
    except:
        #sys.stderr.write("WARNING: Could not set timestamp\n")
        return None

    global dirty
    dirty = True
    return vlan

def get_subnets():
    import pickle
    try:
        subnet_file = open(cache_dir + "/subnets", 'rb')
        subnets = dict(pickle.load(subnet_file))
        subnet_file.close()
    except:
        #sys.stderr.write("WARNING: Could not open subnets cache file. Performing lookup.\n")
        subnets = fetch_subnets()

    #CHECK TIMESTAMP
    try:
        if ( time.time() - subnets['timestamp'] ) > cache_life:
            subnets = fetch_subnets()
    except:
        #sys.stderr.write("WARNING: failed to verify subnet cache timestamp\n")
        subnets = fetch_subnets()

    if dirty:
        #sys.stderr.write("DEBUG: cache dirty, writing back subnets\n")
        subnet_file = open(cache_dir + "/subnets", 'wb')
        pickle.dump(subnets, subnet_file)
        subnet_file.close()

    return subnets

def get_vlans():
    import pickle
    try:
        vlan_file = open(cache_dir + "/vlans", 'rb')
        vlans = dict(pickle.load(vlan_file))
        vlan_file.close()
        #CHECK TIMESTAMPS
        for subnet in vlans.keys():
            try:
                if ( time.time() - vlans[subnet]['timestamp'] ) < cache_life:
                    continue
            except:
                pass #sys.stderr.write("WARNING: failed to verify vlan cache timestamp\n")
                vlan = fetch_vlan(subnet)
                if (vlan == None):
                    #sys.stderr.write("WARNING: vlan fetch returned None\n")
                    continue
                else:
                    vlans[subnet] = vlan
    except:
        #sys.stderr.write("WARNING: Could not open vlans cache file. Performing lookup.\n")
        subnets = get_subnets()['subnets']
        vlans = {}
        for subnet in subnets.keys():
            vlan = fetch_vlan(subnet)
            if (vlan == None):
                pass #sys.stderr.write("WARNING: vlan fetch returned None\n")
                continue
            else:
                vlans[subnet] = vlan

    if dirty:
        #sys.stderr.write("DEBUG: cache dirty, writing back vlans\n")
        vlan_file = open(cache_dir + "/vlans", 'wb')
        pickle.dump(vlans, vlan_file)
        vlan_file.close()

    return vlans

def get_vlan(subnet):
    import pickle
    #sys.stderr.write("DEBUG: getting vlan for subnet " + subnet + ".\n")

    vlans = {}

    try:
        vlan_file = open(cache_dir + "/vlans", 'rb')
        vlans = dict(pickle.load(vlan_file))
        vlan_file.close()
    except:
        #sys.stderr.write("WARNING: Could not open vlans cache file. Performing lookup.\n")
        vlans[subnet] = fetch_vlan(subnet)

    #CHECK TIMESTAMPS
    try:
        if ( time.time() - vlans[subnet]['timestamp'] ) > cache_life:
            vlans[subnet] = fetch_vlan(subnet)
    except:
        #sys.stderr.write("WARNING: failed to verify vlan cache timestamp\n")
        vlans[subnet] = fetch_vlan(subnet)

    try:
        if dirty:
            #sys.stderr.write("DEBUG: cache dirty, writing back vlans\n")
            vlan_file = open(cache_dir + "/vlans", 'wb')
            pickle.dump(vlans, vlan_file)
            vlan_file.close()
    except:
        sys.stderr.write("ERROR: Failed to writeback vlan cache.\n")

    return vlans[subnet]

def get_subnet_id(ip):
    subnets = get_subnets()['subnets']

    for subnet in subnets.keys():
        try:
            if ipaddress.IPAddress(ip) in ipaddress.IPNetwork(subnet):
                return subnet
        except:
            raise
    return None

def get_vlan_id(ip):
    s = get_subnet_id(ip)

    if s == "10.1.152.0/24":
        return 2152
    elif s == "10.1.154.0/24":
        return 2154
    elif s == "10.1.155.0/24":
        return 2155

    if( s != None ):
        try:
            i = get_vlan(s)['id']
            return i
        except:
            pass
    else:
        pass #sys.stderr.write("Got None for subnet")
    return None

def get_address(target):
    try:
        address = ipaddress.ip_address(target)
    except:
        try:
            #assume hostname, do lookup
            ip_string = socket.gethostbyname(target)
            address = ipaddress.ip_address(ip_string)
        except:
            raise
    return address

def slookup(target):
    address = get_address(target)
    subnets = get_subnets()['subnets']
    for subnet in subnets.keys():
        if address in ipaddress.ip_network(subnet):
            return subnet
    return ""

known_vlans = {
        '129.97.15.0/24': '15',
        '10.1.152.0/24': '2152',
        '10.1.154.0/24': '2154',
        '10.1.155.0/24': '2155'
        }

def vlookup(target):
    address = get_address(target)
    subnets = {}
    subnet = {}

    subnets = get_subnets()

    for subnet in subnets['subnets'].keys():
        try:
            if address in ipaddress.ip_network(subnet):
                try:
                    vlan = get_vlan(subnet)
                    if ( vlan == None ):
                        try:
                            return known_vlans[subnet]
                        except:
                            raise "LOOKUP FAILURE"
                    else:
                        return vlan['id']
                except:
                    raise
        except:
            raise
    raise "LOOKUP_FAILURE"

def is_natted(target):
    address = get_address(target)

    prefix = str(address).split('.')[0:2]

    natted_area_one = ipaddress.ip_network("10.0.0.0/12")
    natted_area_two = ipaddress.ip_network("172.19.0.0/16")

    if( (address in natted_area_one) or (address in natted_area_two) ):
        return True
    else:
        return False

def glookup(target):
    address = get_address(target)
    subnets = get_subnets()['subnets']
    for subnet in subnets.keys():
        if address in ipaddress.ip_network(subnet):
            if is_natted(address):
                return str(ipaddress.ip_network(subnet)[1]), str(ipaddress.ip_network(subnet)[-2])
            else:
                return str(ipaddress.ip_network(subnet)[1]), None
    return None, None
